#include "KeyboardSynthesizer.hpp"
#include "KeyBind.hpp"

KeyboardSynthesizer::KeyboardSynthesizer(const KeyBind* keyBind)
: keyBind_(keyBind)
{
	// 擾(0xffff)ŏ
	memset(scanCodes_, 0xff, sizeof(scanCodes_));

	// L[̉
	for (size_t i = 0; i < 256; ++i)
		pressed_[i] = false;

	// L[{[hpf[^
	ZeroMemory(&keyboard_, sizeof(INPUT));
	keyboard_.type = INPUT_KEYBOARD;
}


KeyboardSynthesizer::~KeyboardSynthesizer()
{
	close();
}

bool KeyboardSynthesizer::open()
{
	return true;
}

void KeyboardSynthesizer::close()
{
	reset();
}

void KeyboardSynthesizer::press(BYTE noteNumber)
{
	keyDown(static_cast<BYTE>(keyBind_->getNoteNumber(noteNumber)));
}

void KeyboardSynthesizer::release(BYTE noteNumber)
{
	keyUp(static_cast<BYTE>(keyBind_->getNoteNumber(noteNumber)));
}

void KeyboardSynthesizer::reset()
{
	for (size_t i = 0; i < 256; ++i)
		if (pressed_[i])
			keyUp(static_cast<BYTE>(i));
}

void KeyboardSynthesizer::keyDown(BYTE virtualKey)
{
	WORD scanCode = getScanCode(virtualKey);
	if (scanCode == 0)
		return;

	keyboard_.ki.wScan = scanCode & 0x00ff;

	keyboard_.ki.dwFlags = KEYEVENTF_SCANCODE;
	if ((scanCode & 0xff00) == 0xe000)
		keyboard_.ki.dwFlags |= KEYEVENTF_EXTENDEDKEY;

	SendInput(1, &keyboard_, sizeof(INPUT));
	pressed_[virtualKey] = true;
}

void KeyboardSynthesizer::keyUp(BYTE virtualKey)
{
	WORD scanCode = getScanCode(virtualKey);
	if (scanCode == 0)
		return;

	keyboard_.ki.wScan = scanCode & 0x00ff;

	keyboard_.ki.dwFlags = KEYEVENTF_SCANCODE | KEYEVENTF_KEYUP;
	if ((scanCode & 0xff00) == 0xe000)
		keyboard_.ki.dwFlags |= KEYEVENTF_EXTENDEDKEY;

	SendInput(1, &keyboard_, sizeof(INPUT));
	pressed_[virtualKey] = false;
}

WORD KeyboardSynthesizer::getScanCode(BYTE virtualKey)
{
	if (scanCodes_[virtualKey] == 0xffff)
	{
		WORD code = MapVirtualKey(virtualKey, MAPVK_VK_TO_VSC_EX);
		if ((code & 0xff00) == 0 || (code & 0xff00) == 0xe000)
			scanCodes_[virtualKey] = code;
		else
			scanCodes_[virtualKey] = 0;
	}

	return scanCodes_[virtualKey];
}

