#include "MainWindow.hpp"
#include <Shlwapi.h>
#include "constants.hpp"
#include "KeyBind.hpp"
#include "KeyboardSynthesizer.hpp"
#include "MidiSynthesizer.hpp"
#include "KeySequencer.hpp"
#include "KeyBindWindow.hpp"
#include "resource.h"
#include "StringTable.hpp"
#include "MidiInput.hpp"

MainWindow::MainWindow(HINSTANCE instanceHandle)
: initialized_(false), instanceHandle_(instanceHandle), windowHandle_(nullptr), fileTextHandle_(nullptr), statusTextHandle_(nullptr), openKeyBindButtonHanndle_(nullptr), testPlayButtonHandle_(nullptr), playHotKeyRegistered_(0), midiInputHotKeyRegistered_(0), allStopHotKeyRegistered_(0), keySequencer_(nullptr)
{
	static int staticInitialize = registerWindowClass(instanceHandle);

	keyBind_ = new KeyBind();
	keyboardSynthesizer_ = new KeyboardSynthesizer(keyBind_);
	midiSynthesizer_ = new MidiSynthesizer();
	midiInput_ = new MidiInput(keyboardSynthesizer_);
}

MainWindow::~MainWindow()
{
	delete midiInput_;
	delete midiSynthesizer_;
	delete keyboardSynthesizer_;
	delete keyBind_;
}

bool MainWindow::initialize()
{
	if (initialized_)
		return false;
	initialized_ = true;

	// 쐬EBhẼX^C
	DWORD windowStype = WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX;

	// NCAg̈悩EBhETCYvZ
	int width = CLIENT_WIDTH;
	int height = CLIENT_HEIGHT;
	RECT clientRect = { 0, 0, width, height };
	if (AdjustWindowRect(&clientRect, windowStype, FALSE))
	{
		width = clientRect.right - clientRect.left;
		height = clientRect.bottom - clientRect.top;
	}

	// ʒʒuƂ
	int x = GetSystemMetrics(SM_CXSCREEN);
	int y = GetSystemMetrics(SM_CYSCREEN);
	x = x > width ? (x - width) / 2 : 0;
	y = y > height ? (y - height) / 2 : 0;

	// Window쐬
	windowHandle_ = CreateWindowEx(0, WINDOW_CLASS_NAME, StringTable::get(IDS_MAINWINDOW_CAPTION), windowStype, x, y, width, height, nullptr, nullptr, instanceHandle_, nullptr);
	if (windowHandle_ == nullptr)
	{
		MessageBox(nullptr, StringTable::get(IDS_ERROR_CREATE_WINDOW), StringTable::get(IDS_ERRORDIALOG_CAPTION), ERROR_DIALOG_TYPE);
		return false;
	}

	// Window procedurepthis|C^L
	SetWindowLongPtr(windowHandle_, GWLP_USERDATA, 0);
	if (SetWindowLongPtr(windowHandle_, GWLP_USERDATA, reinterpret_cast<LONG_PTR>(this)) == 0 && GetLastError() != 0)
		goto DESTROY;

	// Đt@C\
	fileTextHandle_ = CreateWindow(_T("STATIC"), StringTable::get(IDS_MAINWINDOW_FILE_EMPTY), WS_VISIBLE | WS_CHILD | WS_BORDER | SS_CENTERIMAGE | SS_CENTER, 8, 8, CLIENT_WIDTH - 16, 30, windowHandle_, nullptr, instanceHandle_, nullptr);
	if (fileTextHandle_ == nullptr)
		goto DESTROY;

	// ĐXe[^X\
	statusTextHandle_ = CreateWindow(_T("STATIC"), _T(""), WS_VISIBLE | WS_CHILD | WS_BORDER | SS_CENTERIMAGE | SS_CENTER, 8, 46, CLIENT_WIDTH - 16, 30, windowHandle_, nullptr, instanceHandle_, nullptr);
	if (statusTextHandle_ == nullptr)
		goto DESTROY;

	// L[oChݒ{^\
	openKeyBindButtonHanndle_ = CreateWindow(_T("BUTTON"), StringTable::get(IDS_MAINWINDOW_KEY_BIND_BUTTON), WS_VISIBLE | WS_CHILD | WS_BORDER | BS_PUSHBUTTON | BS_CENTER, 8, 84, 128, 30, windowHandle_, reinterpret_cast<HMENU>(ID_BUTTON_KEY_BIND), instanceHandle_, nullptr);
	if (openKeyBindButtonHanndle_ == nullptr)
		goto DESTROY;

	// eXgĐ/~{^\
	testPlayButtonHandle_ = CreateWindow(_T("BUTTON"), StringTable::get(IDS_MAINWINDOW_TEST_PLAY_BUTTON), WS_VISIBLE | WS_CHILD | WS_BORDER | BS_PUSHBUTTON | BS_CENTER, 8 + 128 + 8, 84, 192, 30, windowHandle_, reinterpret_cast<HMENU>(ID_BUTTON_TEST_PLAY), instanceHandle_, nullptr);
	if (testPlayButtonHandle_ == nullptr)
		goto DESTROY;

	// Window\
	ShowWindow(windowHandle_, SW_SHOW);

	// Drag & DropL
	DragAcceptFiles(windowHandle_, TRUE);

	// HotkeyL
	RegisterHotKeys();

	// t쐬
	keySequencer_ = new KeySequencer(windowHandle_, keyBind_);
	keySequencer_->threadStart();

	// NɃt@ChbvĂǂݍ
	if (__argc > 1 && __targv[1] != nullptr)
	{
		LRESULT result = keySequencer_->loadFile(__targv[1]);
		if (result == 0)
		{
			// t@C̒擾
			WORD filenameLength = GetFileTitle(__targv[1], nullptr, 0);

			// t@Cp̃m
			TCHAR* filename = new TCHAR[filenameLength]; // ͏I[NULLlengthɓĂ

			// t@C擾
			GetFileTitle(__targv[1], filename, filenameLength);

			// t@CpX̕\؂ւ
			SetWindowText(fileTextHandle_, filename);

			delete[] filename;
		}
	}

	refreshStatus();

	return true;

DESTROY:

	MessageBox(windowHandle_, StringTable::get(IDS_ERROR_CREATE_WINDOW), StringTable::get(IDS_ERRORDIALOG_CAPTION), ERROR_DIALOG_TYPE);
	PostMessage(windowHandle_, WM_CLOSE, 0, 0);
	run();
	return false;
}

void MainWindow::run()
{
	MSG message;
	for (;;)
	{
		BOOL result = GetMessage(&message, nullptr, 0, 0);
		if (result == 0 || result == -1)
			break;
		DispatchMessage(&message);
	}
}

int MainWindow::registerWindowClass(HINSTANCE instanceHandle)
{
	// Ă㏑邾ővȋC邯
	// ꉞŏ1񂾂o^΂Ȃ̂ł̂悤ɏĂ
	// EBhENX̓o^
	WNDCLASSEX wcex;
	wcex.cbSize = sizeof(WNDCLASSEX);
	wcex.lpszClassName = WINDOW_CLASS_NAME;
	wcex.hInstance = instanceHandle;
	wcex.style = 0;
	wcex.cbClsExtra = 0;
	wcex.cbWndExtra = 0;
	wcex.lpszMenuName = nullptr;
	wcex.hbrBackground = BACKGROUND_BRUSH;
	wcex.hIcon = (HCURSOR)LoadImage(instanceHandle, MAKEINTRESOURCE(IDI_ICON1), IMAGE_ICON, 0, 0, LR_DEFAULTSIZE);
	wcex.hIconSm = (HCURSOR)LoadImage(instanceHandle, MAKEINTRESOURCE(IDI_ICON1), IMAGE_ICON, 0, 0, LR_DEFAULTSIZE);
	wcex.hCursor = LoadCursor(NULL, IDC_ARROW);
	wcex.lpfnWndProc = callbackProcedure;
	RegisterClassEx(&wcex);

	return 0;
}

LRESULT WINAPI MainWindow::callbackProcedure(HWND windowHandle, UINT message, WPARAM wParam, LPARAM lParam)
{
	MainWindow* window = reinterpret_cast<MainWindow*>(GetWindowLongPtr(windowHandle, GWLP_USERDATA));
	if (window != nullptr)
		return window->procedure(windowHandle, message, wParam, lParam);

	switch (message)
	{
	case WM_DESTROY:
		PostQuitMessage(0);
		return 0;
	default:
		return DefWindowProc(windowHandle, message, wParam, lParam);
	}
}

LRESULT MainWindow::procedure(HWND windowHandle, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch (message)
	{
	case WM_APP_REFRESH_STATUS:
		refreshStatus();
		break;

	case WM_COMMAND:
		switch (LOWORD(wParam))
		{
		case ID_BUTTON_KEY_BIND:
			{
				// tȂ~߂
				keySequencer_->stop();

				// MIDI͒Ȃ~߂
				midiInput_->close();

				refreshStatus();

				// L[oChݒ_CAON
				KeyBindWindow* child = new KeyBindWindow(instanceHandle_, windowHandle, keyBind_);
				if (child->initialize())
				{
					// ݒ蒆Hotkey͂Ȃ̂ňUHotkey~
					UnregisterHotKeys();

					child->run();

					// Hotkey߂(ύXĂꍇ̍XV˂)
					RegisterHotKeys();
				}
				else
					MessageBox(windowHandle, StringTable::get(IDS_ERROR_OPEN_KEYBINDWINDOW), StringTable::get(IDS_ERRORDIALOG_CAPTION), ERROR_DIALOG_TYPE);
				delete child;
			}
			return 0;

		case ID_BUTTON_TEST_PLAY:
			if (keySequencer_->isPlaying())
			{
				// MIDIeXgĐȂΒ~
				if (keySequencer_->getSynthesizer() == midiSynthesizer_)
					keySequencer_->stop();
				// ĐȂ΍Đ~MIDIeXgĐJn
				else
					keySequencer_->play(midiSynthesizer_);
			}
			// MIDI͎tȂMIDI͂~ăeXgĐJn
			else if (midiInput_->isOpened())
			{
				midiInput_->close();
				keySequencer_->play(midiSynthesizer_);
			}
			// ~Ȃ΍ĐJn
			else
				keySequencer_->play(midiSynthesizer_);

			// \XV
			refreshStatus();
			return 0;
		}
		break;

	case WM_HOTKEY:
		if (wParam == ID_HOTKEY_PLAY)
		{
			if (keySequencer_->isPlaying())
			{
				// MIDIeXgĐȂMIDIeXgĐ~čĐJn
				if (keySequencer_->getSynthesizer() == midiSynthesizer_)
					keySequencer_->play(keyboardSynthesizer_);
				// ĐȂ΍Đ~
				else
					keySequencer_->stop();
			}
			// MIDI͎tȂMIDI͂~čĐJn
			else if (midiInput_->isOpened())
			{
				midiInput_->close();
				keySequencer_->play(keyboardSynthesizer_);
			}
			// ~Ȃ΍ĐJn
			else
				keySequencer_->play(keyboardSynthesizer_);

			// \XV
			refreshStatus();
		}
		else if (wParam == ID_HOTKEY_MIDI_INPUT)
		{
			// Đ or MIDIeXgĐȂΒ~MIDI͂Jn
			if (keySequencer_->isPlaying())
			{
				keySequencer_->stop();

				LRESULT result = midiInput_->open();
				if (result == MidiInput::ERROR_DEVICE_NOT_FOUND)
					MessageBox(windowHandle, StringTable::get(IDS_ERROR_MIDI_INPUT_DEVICE_NOT_FOUND), StringTable::get(IDS_ERRORDIALOG_CAPTION), ERROR_DIALOG_TYPE);
				else if (result == MidiInput::ERROR_UNKNOWN)
					MessageBox(windowHandle, StringTable::get(IDS_ERROR_MIDI_INPUT_UNKNOWN), StringTable::get(IDS_ERRORDIALOG_CAPTION), ERROR_DIALOG_TYPE);
			}
			// MIDI͎tȂΒ~
			else if (midiInput_->isOpened())
				midiInput_->close();
			// ~ȂMIDI͊Jn
			else
			{
				LRESULT result = midiInput_->open();
				if (result == MidiInput::ERROR_DEVICE_NOT_FOUND)
					MessageBox(windowHandle, StringTable::get(IDS_ERROR_MIDI_INPUT_DEVICE_NOT_FOUND), StringTable::get(IDS_ERRORDIALOG_CAPTION), ERROR_DIALOG_TYPE);
				else if (result == MidiInput::ERROR_UNKNOWN)
					MessageBox(windowHandle, StringTable::get(IDS_ERROR_MIDI_INPUT_UNKNOWN), StringTable::get(IDS_ERRORDIALOG_CAPTION), ERROR_DIALOG_TYPE);
			}

			// \XV
			refreshStatus();
		}
		else if (wParam == ID_HOTKEY_ALL_STOP)
		{
			// Đ or MIDIeXgĐȂΒ~
			if (keySequencer_->isPlaying())
			{
				keySequencer_->stop();

				// \XV
				refreshStatus();
			}
			// MIDI͎tȂΒ~
			else if (midiInput_->isOpened())
			{
				midiInput_->close();

				// \XV
				refreshStatus();
			}
		}
		return 0;

	case WM_DROPFILES:
		{
			HDROP handle = reinterpret_cast<HDROP>(wParam);

			// hbvꂽt@C(1)̃pX̒擾
			UINT filepathLength = DragQueryFile(handle, 0, nullptr, 0);

			// t@CpXp̃m
			TCHAR* filepath = new TCHAR[filepathLength + 1]; // I[NULLlengthɓȂ

			// t@CpX擾
			DragQueryFile(handle, 0, filepath, filepathLength + 1);
			DragFinish(handle);

			// tȂ~߂
			keySequencer_->stop();

			// MIDI͒Ȃ~߂
			midiInput_->close();

			refreshStatus();

			// t@Cǂݍ
			LRESULT result = keySequencer_->loadFile(filepath);
			if (result == 0)
			{
				// t@C̒擾
				WORD filenameLength = GetFileTitle(filepath, nullptr, 0);

				// t@Cp̃m
				TCHAR* filename = new TCHAR[filenameLength]; // ͏I[NULLlengthɓĂ

				// t@C擾
				GetFileTitle(filepath, filename, filenameLength);

				// t@CpX̕\؂ւ
				SetWindowText(fileTextHandle_, filename);

				delete[] filename;
			}
			else
			{
				SetWindowText(fileTextHandle_, StringTable::get(IDS_MAINWINDOW_FILE_EMPTY));

				switch (result)
				{
				case StandardMidiFormat::LOAD_ERROR_FILE_OPEN:
					// t@CI[voȂ
					MessageBox(windowHandle, StringTable::get(IDS_ERROR_OPEN_FILE), StringTable::get(IDS_ERRORDIALOG_CAPTION), ERROR_DIALOG_TYPE);
					break;
				case StandardMidiFormat::LOAD_ERROR_FILE_TOO_LARGE:
					// 4GBȏ̃t@C͔Ή
					MessageBox(windowHandle, StringTable::get(IDS_ERROR_TOO_LARGE_FILE), StringTable::get(IDS_ERRORDIALOG_CAPTION), ERROR_DIALOG_TYPE);
					break;
				case StandardMidiFormat::LOAD_ERROR_FILE_READ:
					// t@CǂݍݒɕsȃG[
					MessageBox(windowHandle, StringTable::get(IDS_ERROR_READ_FILE), StringTable::get(IDS_ERRORDIALOG_CAPTION), ERROR_DIALOG_TYPE);
					break;
				case StandardMidiFormat::LOAD_ERROR_ILLEGAL_SMF:
					// SMFƂĕs
					MessageBox(windowHandle, StringTable::get(IDS_ERROR_ILLEGAL_SMF_FILE), StringTable::get(IDS_ERRORDIALOG_CAPTION), ERROR_DIALOG_TYPE);
					break;
				case StandardMidiFormat::LOAD_ERROR_NOT_SUPPORT_SMF:
					// T|[gĂȂSMF`
					MessageBox(windowHandle, StringTable::get(IDS_ERROR_UNSUPPORT_SMF_FILE), StringTable::get(IDS_ERRORDIALOG_CAPTION), ERROR_DIALOG_TYPE);
					break;
				}
			}

			delete[] filepath;
		}
		return 0;

	case WM_CTLCOLORSTATIC:
		SetBkMode(reinterpret_cast<HDC>(wParam), TRANSPARENT);
		return reinterpret_cast<LRESULT>(BACKGROUND_BRUSH);

	case WM_CLOSE:
		// EBhEQƂ̂ɉĂDestroyWindow
		UnregisterHotKeys();

		if (keySequencer_ != nullptr)
		{
			delete keySequencer_;
			keySequencer_ = nullptr;
		}

		DestroyWindow(windowHandle);
		return 0;

	case WM_DESTROY:
		PostQuitMessage(0);
		return 0;
	}

	return DefWindowProc(windowHandle, message, wParam, lParam);
}

void MainWindow::RegisterHotKeys()
{
	WORD hotkey;

	hotkey = keyBind_->getPlayHotKey();
	if (playHotKeyRegistered_ != hotkey)
	{
		if (playHotKeyRegistered_ != 0)
		{
			UnregisterHotKey(windowHandle_, ID_HOTKEY_PLAY);
			playHotKeyRegistered_ = 0;
		}

		if (hotkey != 0)
		{
			UINT modifiers = 0;
			if (hotkey & KeyBind::ModifierKey::SHIFT)
				modifiers |= MOD_SHIFT;
			if (hotkey & KeyBind::ModifierKey::CONTROL)
				modifiers |= MOD_CONTROL;
			if (hotkey & KeyBind::ModifierKey::ALTERNATE)
				modifiers |= MOD_ALT;

			if (RegisterHotKey(windowHandle_, ID_HOTKEY_PLAY, modifiers, hotkey & 0x00ff))
				playHotKeyRegistered_ = hotkey;
			else
				MessageBox(windowHandle_, StringTable::get(IDS_ERROR_REGISTER_PLAY_HOTKEY), StringTable::get(IDS_ERRORDIALOG_CAPTION), ERROR_DIALOG_TYPE);
		}
	}

	hotkey = keyBind_->getMidiInputHotKey();
	if (midiInputHotKeyRegistered_ != hotkey)
	{
		if (midiInputHotKeyRegistered_ != 0)
		{
			UnregisterHotKey(windowHandle_, ID_HOTKEY_MIDI_INPUT);
			midiInputHotKeyRegistered_ = 0;
		}

		if (hotkey != 0)
		{
			UINT modifiers = 0;
			if (hotkey & KeyBind::ModifierKey::SHIFT)
				modifiers |= MOD_SHIFT;
			if (hotkey & KeyBind::ModifierKey::CONTROL)
				modifiers |= MOD_CONTROL;
			if (hotkey & KeyBind::ModifierKey::ALTERNATE)
				modifiers |= MOD_ALT;

			if (RegisterHotKey(windowHandle_, ID_HOTKEY_MIDI_INPUT, modifiers, hotkey & 0x00ff))
				midiInputHotKeyRegistered_ = hotkey;
			else
				MessageBox(windowHandle_, StringTable::get(IDS_ERROR_REGISTER_MIDI_INPUT_HOTKEY), StringTable::get(IDS_ERRORDIALOG_CAPTION), ERROR_DIALOG_TYPE);
		}
	}

	hotkey = keyBind_->getAllStopHotKey();
	if (allStopHotKeyRegistered_ != hotkey)
	{
		if (allStopHotKeyRegistered_ != 0)
		{
			UnregisterHotKey(windowHandle_, ID_HOTKEY_ALL_STOP);
			allStopHotKeyRegistered_ = 0;
		}

		if (hotkey != 0)
		{
			UINT modifiers = 0;
			if (hotkey & KeyBind::ModifierKey::SHIFT)
				modifiers |= MOD_SHIFT;
			if (hotkey & KeyBind::ModifierKey::CONTROL)
				modifiers |= MOD_CONTROL;
			if (hotkey & KeyBind::ModifierKey::ALTERNATE)
				modifiers |= MOD_ALT;

			if (RegisterHotKey(windowHandle_, ID_HOTKEY_ALL_STOP, modifiers, hotkey & 0x00ff))
				allStopHotKeyRegistered_ = hotkey;
			else
				MessageBox(windowHandle_, StringTable::get(IDS_ERROR_REGISTER_ALL_STOP_HOTKEY), StringTable::get(IDS_ERRORDIALOG_CAPTION), ERROR_DIALOG_TYPE);
		}
	}
}

void MainWindow::UnregisterHotKeys()
{
	// ɋCɂUnregisterHotKeyAłĂvȋC邯ǈꉞ䂷

	if (allStopHotKeyRegistered_ != 0)
	{
		UnregisterHotKey(windowHandle_, ID_HOTKEY_ALL_STOP);
		allStopHotKeyRegistered_ = 0;
	}

	if (midiInputHotKeyRegistered_ != 0)
	{
		UnregisterHotKey(windowHandle_, ID_HOTKEY_MIDI_INPUT);
		midiInputHotKeyRegistered_ = 0;
	}

	if (playHotKeyRegistered_ != 0)
	{
		UnregisterHotKey(windowHandle_, ID_HOTKEY_PLAY);
		playHotKeyRegistered_ = 0;
	}
}

void MainWindow::refreshStatus()
{
	// MIDIeXgĐ{^̕\
	if (keySequencer_->isPlaying() && keySequencer_->getSynthesizer() == midiSynthesizer_)
		SetWindowText(testPlayButtonHandle_, StringTable::get(IDS_MAINWINDOW_STOP_BUTTON));
	else
		SetWindowText(testPlayButtonHandle_, StringTable::get(IDS_MAINWINDOW_TEST_PLAY_BUTTON));

	// Xe[^X\
	if (keySequencer_->isPlaying())
	{
		// MIDIeXgĐ
		if (keySequencer_->getSynthesizer() == midiSynthesizer_)
			SetWindowText(statusTextHandle_, StringTable::get(IDS_MAINWINDOW_STATUS_MIDI_FILE_TEST_PLAYING));
		// Đ
		else
			SetWindowText(statusTextHandle_, StringTable::get(IDS_MAINWINDOW_STATUS_PLAYING));
	}
	// MIDI͎t
	else if (midiInput_->isOpened())
		SetWindowText(statusTextHandle_, StringTable::get(IDS_MAINWINDOW_STATUS_MIDI_INPUTING));
	// ~
	else
		SetWindowText(statusTextHandle_, StringTable::get(IDS_MAINWINDOW_STATUS_STOPPED));
}
