#include "MidiSynthesizer.hpp"
#include "constants.hpp"
#include "resource.h"
#include "StringTable.hpp"

MidiSynthesizer::~MidiSynthesizer()
{
	if (midiOutHandle_ != nullptr)
	{
		midiOutReset(midiOutHandle_);
		midiOutClose(midiOutHandle_);
	}
}

bool MidiSynthesizer::open()
{
	if (midiOutHandle_ != nullptr)
		return true;

	if (midiOutOpen(&midiOutHandle_, MIDIMAPPER, 0, 0, CALLBACK_NULL) != MMSYSERR_NOERROR)
	{
		midiOutHandle_ = nullptr;
		return false;
	}
	else
		return true;
}

void MidiSynthesizer::close()
{
	if (midiOutHandle_ != nullptr)
	{
		midiOutReset(midiOutHandle_);
		midiOutClose(midiOutHandle_);
		midiOutHandle_ = nullptr;
	}
}

void MidiSynthesizer::press(BYTE noteNumber)
{
	if (midiOutHandle_ != nullptr)
		midiOutShortMsg(midiOutHandle_, 0x007f0090 | (static_cast<DWORD>(noteNumber) << 8));
}

void MidiSynthesizer::release(BYTE noteNumber)
{
	if (midiOutHandle_ != nullptr)
		midiOutShortMsg(midiOutHandle_, 0x00000080 | (static_cast<DWORD>(noteNumber) << 8));
}

void MidiSynthesizer::reset()
{
	if (midiOutHandle_ != nullptr)
		midiOutReset(midiOutHandle_);
}

