#include "StandardMidiFormat.hpp"

StandardMidiFormat::StandardMidiFormat()
: fileLoaded_(false), dataLength_(0), data_(nullptr), trackReaders_(nullptr)
{
}

StandardMidiFormat::~StandardMidiFormat()
{
	if (trackReaders_ != nullptr)
		delete[] trackReaders_;

	if (data_ != nullptr)
		delete[] data_;
}

LRESULT StandardMidiFormat::loadFile(const TCHAR* filepath)
{
	// ݓǂݍłf[^j
	fileLoaded_ = false;
	if (data_ != nullptr)
	{
		delete[] data_;
		data_ = nullptr;
		dataLength_ = 0;
	}
	if (trackReaders_ != nullptr)
	{
		delete[] trackReaders_;
		trackReaders_ = nullptr;
	}

	// t@CI[v
	HANDLE handle = CreateFile(filepath, GENERIC_READ, FILE_SHARE_READ, nullptr, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, nullptr);
	if (handle == INVALID_HANDLE_VALUE)
		return LOAD_ERROR_FILE_OPEN;

	// t@CTCY擾
	LARGE_INTEGER filesize;
	if (!GetFileSizeEx(handle, &filesize))
	{
		CloseHandle(handle);
		return LOAD_ERROR_FILE_OPEN; // TCYȂƂI[vsł
	}

	// 4GBȏ̃t@Cɂ͔Ή
	// 32bitłƂƏTCYłmۃG[ɂȂ邯
	if (filesize.QuadPart > 0x00000000ffffffff)
	{
		CloseHandle(handle);
		return LOAD_ERROR_FILE_TOO_LARGE;
	}

	// wb_14byte͍ŒᖳSMFƂĂ)
	if (filesize.QuadPart < 14)
	{
		CloseHandle(handle);
		return LOAD_ERROR_ILLEGAL_SMF;
	}

	dataLength_ = static_cast<DWORD>(filesize.LowPart);


	// m
	data_ = new BYTE[dataLength_];


	// t@Cǂݍ
	DWORD read;
	BOOL result = ReadFile(handle, data_, dataLength_, &read, nullptr);
	CloseHandle(handle);
	if (!result || dataLength_ != read)
		return LOAD_ERROR_FILE_READ;

	BYTE* cursor = data_;
	BYTE* end = cursor + dataLength_;


	// MThd̕
	if (readDword(cursor) != CHUNK_TYPE_HEADER)
		return LOAD_ERROR_ILLEGAL_SMF;

	// wb_`N̑傫(6Œ)
	if (readDword(cursor) != HEADER_CHUNK_LENGTH)
		return LOAD_ERROR_ILLEGAL_SMF;

	// tH[}bgmF
	WORD format = readWord(cursor);
	if (format == Format::TYPE_0)
		format_ = Format::TYPE_0;
	else if (format == Format::TYPE_1)
		format_ = Format::TYPE_1;
	else
		return LOAD_ERROR_NOT_SUPPORT_SMF;

	// gbNǂݎ
	track_ = readWord(cursor);
	trackReaders_ = new TrackReader[track_];

	// ^Cx[Xǂݎ
	quarterNoteTick_ = readWord(cursor);
	if ((quarterNoteTick_ & 0x8000) != 0)
		return LOAD_ERROR_NOT_SUPPORT_SMF;

	for (size_t track = 0; track < track_; ++track)
	{
		if (cursor + 8 > end)
			return LOAD_ERROR_ILLEGAL_SMF;

		// MTrk̕
		if (cursor + 4 <= end && readDword(cursor) != CHUNK_TYPE_TRACK)
			return LOAD_ERROR_ILLEGAL_SMF;

		// gbNf[^̓ǂݎ
		DWORD trackLength = readDword(cursor);

		// gbNf[^XLbv
		cursor += trackLength;
		if (cursor > end)
			return LOAD_ERROR_ILLEGAL_SMF;

		// gbNf[^̈ʒuL^
		trackReaders_[track].start = cursor - trackLength;
		trackReaders_[track].end = cursor;
	}

	omni_ = false;
	mono_ = false;

	for (size_t track = 0; track < track_; ++track)
	{
		// J[\擪ɃZbg
		trackReaders_[track].cursor = trackReaders_[track].start;

		// ŏ̃f^^Cǂݍ
		DWORD deltaTime = 0;
		if (trackReaders_[track].start < trackReaders_[track].end)
		{
			for (size_t i = 0; i <= 4; ++i)
			{
				if (i == 4 || trackReaders_[track].cursor >= trackReaders_[track].end)
					return LOAD_ERROR_ILLEGAL_SMF;

				deltaTime |= (*trackReaders_[track].cursor & 0x7f);
				if ((*trackReaders_[track].cursor & 0x80) == 0)
				{
					++trackReaders_[track].cursor;
					break;
				}
				else
				{
					deltaTime <<= 7;
					++trackReaders_[track].cursor;
				}
			}
		}

	}

	for (;;)
	{
		size_t minimumDeltaTimeTrack = 0;
		DWORD minimumDeltaTime = 0;
		
		// CxgcĂgbN擾
		size_t track = 0;
		for (; track < track_; ++track)
		{
			if (trackReaders_[track].cursor < trackReaders_[track].end)
			{
				minimumDeltaTimeTrack = track;
				minimumDeltaTime = trackReaders_[track].deltaTime;
				break;
			}
		}


		// CxgcĂȂΏI
		if (track >= track_)
			break;

		// ɃCxg̎cĂgbN΂̒f^^Cŏ̕I
		// (f^^CȂgbNԍႢI)
		for (; track < track_; ++track)
		{
			if (trackReaders_[track].cursor < trackReaders_[track].end)
			{
				if (trackReaders_[track].deltaTime < minimumDeltaTime)
				{
					minimumDeltaTimeTrack = track;
					minimumDeltaTime = trackReaders_[track].deltaTime;
				}
			}
		}

		// f^^Cŏ̃gbN烁bZ[Wǂݎ
		TrackReader* reader = &trackReaders_[minimumDeltaTimeTrack];
		BYTE* cursor = reader->cursor;

		// jOXe[^XlăXe[^XoCgǂݎ
		BYTE status = *cursor;
		if ((status & 0x80) == 0)
		{
			status = reader->lastMidiEvent;
			--cursor; // Xe[^XoCgLƂƖœolA-1ĕ␳
		}

		// ׂ̏̈ɋL
		if (status < 0xf0)
			reader->lastMidiEvent = status;

		size_t eventLength = 0;

		switch (status & 0xf0)
		{
		case 0xd0: // `lvbV[
		case 0xc0: // vO`FW
			eventLength = 2;
			break;

		case 0x80: // m[g OFF
		case 0x90: // m[g ON
		case 0xa0: // |tHjbNL[vbV[
		case 0xe0: // sb`xh
			eventLength = 3;
			break;

		// Rg[`FW or `l[hbZ[W
		case 0xb0:
			eventLength = 3;

			// OMNI OFF
			if (cursor[1] == 0x7c)
				omni_ = false;
			// OMNI ON
			else if (cursor[1] == 0x7d)
				omni_ = true;
			// POLY[h
			else if (cursor[1] == 0x7e)
				mono_ = false;
			// MONO[h
			else if (cursor[1] == 0x7f)
			{
				mono_ = true;
				if (!omni_)
					++eventLength; // ̃P[XCxg̃TCY4byteɂȂ
			}
			break;

		case 0xf0:
			// SysExCxg (F0)
			if (status == 0xf0 || status == 0xf7)
			{
				eventLength = static_cast<WORD>(cursor[1]) + 2;
			}
			// ^Cxg
			else if (status == 0xff)
			{
				eventLength = static_cast<WORD>(cursor[2]) + 3;
			}
			break;
		}

		if (eventLength == 0)
			return LOAD_ERROR_ILLEGAL_SMF;

		cursor += eventLength;
		if (cursor > reader->end)
			return LOAD_ERROR_ILLEGAL_SMF;

		// I[łȂΎ̃bZ[W̃f^^CǂݎĂ
		DWORD deltaTime = 0;
		if (cursor < reader->end)
		{
			for (size_t i = 0; i <= 4; ++i)
			{
				if (i == 4 || cursor >= reader->end)
					return LOAD_ERROR_ILLEGAL_SMF;

				deltaTime |= (*cursor & 0x7f);
				if ((*cursor & 0x80) == 0)
				{
					++cursor;
					break;
				}
				else
				{
					deltaTime <<= 7;
					++cursor;
				}
			}
		}

		// ꎞϐ̃f[^𔽉f
		reader->cursor = cursor;
		reader->deltaTime = deltaTime;
	}

	fileLoaded_ = true;
	return 0;
}

bool StandardMidiFormat::next(StandardMidiFormatEvent* event)
{
	size_t minimumDeltaTimeTrack = 0;
	DWORD minimumDeltaTime = 0;
	
	// CxgcĂgbN擾
	size_t track = 0;
	for (; track < track_; ++track)
	{
		if (trackReaders_[track].cursor < trackReaders_[track].end)
		{
			minimumDeltaTimeTrack = track;
			minimumDeltaTime = trackReaders_[track].deltaTime;
			break;
		}
	}

	// CxgcĂȂΏI
	if (track >= track_)
		return false;

	// ɃCxg̎cĂgbN΂̒f^^Cŏ̕I
	// (f^^CȂgbNԍႢI)
	for (; track < track_; ++track)
	{
		if (trackReaders_[track].cursor < trackReaders_[track].end)
		{
			if (trackReaders_[track].deltaTime < minimumDeltaTime)
			{
				minimumDeltaTimeTrack = track;
				minimumDeltaTime = trackReaders_[track].deltaTime;
			}
		}
	}

	// f^^Cŏ̃gbN烁bZ[Wǂݎ
	TrackReader* reader = &trackReaders_[minimumDeltaTimeTrack];
	BYTE* cursor = reader->cursor;

	// f^^Co
	event->deltaTime = minimumDeltaTime;

	// jOXe[^XlăXe[^XoCgǂݎ
	BYTE status = *cursor;
	if ((status & 0x80) == 0)
	{
		status = reader->lastMidiEvent;
		--cursor; // Xe[^XoCgLƂƖœolA-1ĕ␳
	}

	// ׂ̏̈ɋL
	if (status < 0xf0)
		reader->lastMidiEvent = status;

	switch (status & 0xf0)
	{
	// m[g OFF
	case 0x80:
		event->type = StandardMidiFormatEvent::Type::MIDI;
		event->midiType = StandardMidiFormatEvent::MidiType::NOTE_OFF;
		event->channel = (status & 0x0f);
		event->noteNumber = cursor[1];
		event->velocity = cursor[2];
		cursor += 3;
		break;

	// m[g ON
	case 0x90:
		event->type = StandardMidiFormatEvent::Type::MIDI;
		event->midiType = StandardMidiFormatEvent::MidiType::NOTE_ON;
		event->channel = (status & 0x0f);
		event->noteNumber = cursor[1];
		event->velocity = cursor[2];
		cursor += 3;
		break;

	// |tHjbNL[vbV[
	case 0xa0:
		event->type = StandardMidiFormatEvent::Type::MIDI;
		event->midiType = StandardMidiFormatEvent::MidiType::POLYPHONIC_KEY_PRESSURE;
		event->channel = (status & 0x0f);
		event->noteNumber = cursor[1];
		event->velocity = cursor[2];
		cursor += 3;
		break;

	// Rg[`FW or `l[hbZ[W
	case 0xb0:
		event->type = StandardMidiFormatEvent::Type::MIDI;
		event->midiType = StandardMidiFormatEvent::MidiType::CONTROL_CHANGE;
		event->channel = (status & 0x0f);
		event->controlNumber = cursor[1];
		event->controlValue = cursor[2];

		// OMNI OFF
		if (cursor[1] == 0x7c)
		{
			omni_ = false;
		}
		// OMNI ON
		else if (cursor[1] == 0x7d)
		{
			omni_ = true;
		}
		// POLY[h
		else if (cursor[1] == 0x7e)
		{
			mono_ = false;
		}
		// MONO[h
		else if (cursor[1] == 0x7f)
		{
			mono_ = true;
			if (!omni_)
				++cursor; // ̃P[XCxg̃TCY4byteɂȂ
		}

		cursor += 3;
		break;

	// vO`FW
	case 0xc0:
		event->type = StandardMidiFormatEvent::Type::MIDI;
		event->midiType = StandardMidiFormatEvent::MidiType::PROGRAM_CHANGE;
		event->channel = (status & 0x0f);
		event->program = cursor[1];
		cursor += 2;
		break;

	// `lvbV[
	case 0xd0:
		event->type = StandardMidiFormatEvent::Type::MIDI;
		event->midiType = StandardMidiFormatEvent::MidiType::CHANNEL_PRESSURE;
		event->channel = (status & 0x0f);
		event->velocity = cursor[1];
		cursor += 2;
		break;

	// sb`xh
	case 0xe0:
		event->type = StandardMidiFormatEvent::Type::MIDI;
		event->midiType = StandardMidiFormatEvent::MidiType::PITCH_BEND;
		event->pitchBend = static_cast<WORD>(cursor[1] & 0x7f) | (static_cast<WORD>(cursor[2] & 0x7f) << 7);
		cursor += 3;
		break;

	case 0xf0:
		// SysExCxg (F0)
		if (status == 0xf0)
		{
			event->type = StandardMidiFormatEvent::Type::SYS_EX;
			event->sysExType = StandardMidiFormatEvent::SysExType::F0;
			event->dataLength = static_cast<WORD>(cursor[1]) + 2;
			event->data = cursor;
			cursor += event->dataLength;
		}
		// SysExCxg (F7)
		else if (status == 0xf7)
		{
			event->type = StandardMidiFormatEvent::Type::SYS_EX;
			event->sysExType = StandardMidiFormatEvent::SysExType::F7;
			event->dataLength = static_cast<WORD>(cursor[1]);
			event->data = &cursor[2];
			cursor += event->dataLength + 2;
		}
		// ^Cxg
		else if (status == 0xff)
		{
			event->type = StandardMidiFormatEvent::Type::META;
			switch (cursor[1])
			{
			case StandardMidiFormatEvent::SEQUENCE_NUMBER: event->metaType = StandardMidiFormatEvent::SEQUENCE_NUMBER; break;
			case StandardMidiFormatEvent::TEXT: event->metaType = StandardMidiFormatEvent::TEXT; break;
			case StandardMidiFormatEvent::COPYRIGHT: event->metaType = StandardMidiFormatEvent::COPYRIGHT; break;
			case StandardMidiFormatEvent::SEQUENCE: event->metaType = StandardMidiFormatEvent::SEQUENCE; break;
			case StandardMidiFormatEvent::INSTRUMENT_NAME: event->metaType = StandardMidiFormatEvent::INSTRUMENT_NAME; break;
			case StandardMidiFormatEvent::LYRIC: event->metaType = StandardMidiFormatEvent::LYRIC; break;
			case StandardMidiFormatEvent::MARKER: event->metaType = StandardMidiFormatEvent::MARKER; break;
			case StandardMidiFormatEvent::CUE_POINT: event->metaType = StandardMidiFormatEvent::CUE_POINT; break;
			case StandardMidiFormatEvent::PROGRAM_NAME: event->metaType = StandardMidiFormatEvent::PROGRAM_NAME; break;
			case StandardMidiFormatEvent::DEVICE_NAME: event->metaType = StandardMidiFormatEvent::DEVICE_NAME; break;
			case StandardMidiFormatEvent::MIDI_CHANNEL_PREFIX: event->metaType = StandardMidiFormatEvent::MIDI_CHANNEL_PREFIX; break;
			case StandardMidiFormatEvent::MIDI_PORT: event->metaType = StandardMidiFormatEvent::MIDI_PORT; break;
			case StandardMidiFormatEvent::END_OF_TRACK: event->metaType = StandardMidiFormatEvent::END_OF_TRACK; break;
			case StandardMidiFormatEvent::TEMPO: event->metaType = StandardMidiFormatEvent::TEMPO; break;
			case StandardMidiFormatEvent::SMPTE_OFFSET: event->metaType = StandardMidiFormatEvent::SMPTE_OFFSET; break;
			case StandardMidiFormatEvent::TIME_SIGNATURE: event->metaType = StandardMidiFormatEvent::TIME_SIGNATURE; break;
			case StandardMidiFormatEvent::KEY_SIGNATURE: event->metaType = StandardMidiFormatEvent::KEY_SIGNATURE; break;
			case StandardMidiFormatEvent::SEQUENCER_SPECIFIC_EVENT: event->metaType = StandardMidiFormatEvent::SEQUENCER_SPECIFIC_EVENT; break;
			}
			event->dataLength = static_cast<WORD>(cursor[2]);
			event->data = &cursor[3];
			cursor += event->dataLength + 3;
		}
		break;
	}

	// I[łȂΎ̃bZ[W̃f^^CǂݎĂ
	DWORD deltaTime = 0;
	if (cursor < reader->end)
	{
		for (size_t i = 0; i < 4; ++i)
		{
			deltaTime |= (*cursor & 0x7f);
			if ((*cursor & 0x80) == 0)
				break;
			else
			{
				deltaTime <<= 7;
				++cursor;
			}
		}
		++cursor;
	}

	// ꎞϐ̃f[^𔽉f
	reader->cursor = cursor;
	reader->deltaTime = deltaTime;

	return true;
}

void StandardMidiFormat::rewind()
{
	omni_ = false;
	mono_ = false;

	for (size_t track = 0; track < track_; ++track)
	{
		trackReaders_[track].cursor = trackReaders_[track].start;

		DWORD deltaTime = 0;
		if (trackReaders_[track].cursor < trackReaders_[track].end)
		{
			for (size_t i = 0; i < 4; ++i)
			{
				deltaTime |= (*trackReaders_[track].cursor & 0x7f);
				if ((*trackReaders_[track].cursor & 0x80) == 0)
					break;
				else
				{
					deltaTime <<= 7;
					++trackReaders_[track].cursor;
				}
			}
			++trackReaders_[track].cursor;
		}
	}
}
